# -*- coding: utf-8 -*-

"""Provide regulatory on/off minimization (ROOM)."""

from __future__ import absolute_import, division

from optlang.symbolics import Zero

from cobra.flux_analysis.parsimonious import pfba


def room(model, solution=None, linear=False, delta=0.03, epsilon=1E-03):
    """
    Compute a single solution based on regulatory on/off minimization (ROOM).

    Compute a new flux distribution that minimizes the number of active
    reactions needed to accommodate a previous reference solution.
    Regulatory on/off minimization (ROOM) is generally used to assess the
    impact of knock-outs. Thus the typical usage is to provide a wildtype flux
    distribution as reference and a model in knock-out state.

    Parameters
    ----------
    model : cobra.Model
        The model state to compute a ROOM-based solution for.
    solution : cobra.Solution, optional
        A (wildtype) reference solution.
    linear : bool, optional
        Whether to use the linear ROOM formulation or not (default False).
    delta: float, optional
        The relative tolerance range (additive) (default 0.03).
    epsilon: float, optional
        The absolute tolerance range (multiplicative) (default 0.001).

    Returns
    -------
    cobra.Solution
        A flux distribution with minimal active reaction changes compared to
        the reference.

    See Also
    --------
    add_room : add ROOM constraints and objective

    """
    with model:
        add_room(model=model, solution=solution, linear=linear, delta=delta,
                 epsilon=epsilon)
        solution = model.optimize()
    return solution


def add_room(model, solution=None, linear=False, delta=0.03, epsilon=1E-03):
    r"""
    Add constraints and objective for ROOM.

    This function adds variables and constraints for applying regulatory
    on/off minimization (ROOM) to the model.

    Parameters
    ----------
    model : cobra.Model
        The model to add ROOM constraints and objective to.
    solution : cobra.Solution, optional
        A previous solution to use as a reference. If no solution is given,
        one will be computed using pFBA.
    linear : bool, optional
        Whether to use the linear ROOM formulation or not (default False).
    delta: float, optional
        The relative tolerance range which is additive in nature
        (default 0.03).
    epsilon: float, optional
        The absolute range of tolerance which is multiplicative
        (default 0.001).

    Notes
    -----
    The formulation used here is the same as stated in the original paper [1]_.
    The mathematical expression is given below:

    minimize \sum_{i=1}^m y^i
    s.t. Sv = 0
         v_min <= v <= v_max
         v_j = 0
         j ∈ A
         for 1 <= i <= m
         v_i - y_i(v_{max,i} - w_i^u) <= w_i^u      (1)
         v_i - y_i(v_{min,i} - w_i^l) <= w_i^l      (2)
         y_i ∈ {0,1}                                (3)
         w_i^u = w_i + \delta|w_i| + \epsilon
         w_i^l = w_i - \delta|w_i| - \epsilon

    So, for the linear version of the ROOM , constraint (3) is relaxed to
    0 <= y_i <= 1.

    See Also
    --------
    pfba : parsimonious FBA

    References
    ----------
    .. [1] Tomer Shlomi, Omer Berkman and Eytan Ruppin, "Regulatory on/off
     minimization of metabolic flux changes after genetic perturbations",
     PNAS 2005 102 (21) 7695-7700; doi:10.1073/pnas.0406346102

    """

    if 'room_old_objective' in model.solver.variables:
        raise ValueError('model is already adjusted for ROOM')

    # optimizes if no reference solution is provided
    if solution is None:
        solution = pfba(model)

    prob = model.problem
    variable = prob.Variable("room_old_objective", ub=solution.objective_value)
    constraint = prob.Constraint(
        model.solver.objective.expression - variable,
        ub=0.0,
        lb=0.0,
        name="room_old_objective_constraint"
    )
    model.objective = prob.Objective(Zero, direction="min", sloppy=True)
    vars_and_cons = [variable, constraint]
    obj_vars = []

    for rxn in model.reactions:
        flux = solution.fluxes[rxn.id]

        if linear:
            y = prob.Variable("y_" + rxn.id, lb=0, ub=1)
            delta = epsilon = 0.0
        else:
            y = prob.Variable("y_" + rxn.id, type="binary")

        # upper constraint
        w_u = flux + (delta * abs(flux)) + epsilon
        upper_const = prob.Constraint(
            rxn.flux_expression - y * (rxn.upper_bound - w_u),
            ub=w_u, name="room_constraint_upper_" + rxn.id)
        # lower constraint
        w_l = flux - (delta * abs(flux)) - epsilon
        lower_const = prob.Constraint(
            rxn.flux_expression - y * (rxn.lower_bound - w_l),
            lb=w_l, name="room_constraint_lower_" + rxn.id)
        vars_and_cons.extend([y, upper_const, lower_const])
        obj_vars.append(y)

    model.add_cons_vars(vars_and_cons)
    model.objective.set_linear_coefficients({v: 1.0 for v in obj_vars})



# -*- coding: utf-8 -*-

"""Test functions of model.py"""


import pytest

from cobra.core import Group


def test_group_add_elements(model):
    num_members = 5
    reactions_for_group = model.reactions[0:num_members]
    group = Group("arbitrary_group1")
    group.add_members(reactions_for_group)
    group.kind = "collection"
    # number of member sin group should equal the number of reactions
    # assigned to the group
    assert len(group.members) == num_members

    # Choose an overlapping, larger subset of reactions for the group
    num_total_members = 12
    reactions_for_larger_group = model.reactions[0:num_total_members]
    group.add_members(reactions_for_larger_group)
    assert len(group.members) == num_total_members


def test_group_kind():
    group = Group("arbitrary_group1")
    with pytest.raises(ValueError) as excinfo:
        group.kind = "non-SBML compliant group kind"
    assert "Kind can only by one of:" in str(excinfo.value)

    group.kind = "collection"
    assert group.kind == "collection"



# -*- coding: utf-8 -*-

"""Module implementing flux sampling for cobra models."""

from __future__ import absolute_import, division

import pandas

from cobra.sampling.achr import ACHRSampler
from cobra.sampling.optgp import OptGPSampler


def sample(model, n, method="optgp", thinning=100, processes=1, seed=None):
    """Sample valid flux distributions from a cobra model.

    The function samples valid flux distributions from a cobra model.
    Currently we support two methods:

    1. 'optgp' (default) which uses the OptGPSampler that supports parallel
        sampling [1]_. Requires large numbers of samples to be performant
        (n < 1000). For smaller samples 'achr' might be better suited.

    or

    2. 'achr' which uses artificial centering hit-and-run. This is a single
       process method with good convergence [2]_.

    Parameters
    ----------
    model : cobra.Model
        The model from which to sample flux distributions.
    n : int
        The number of samples to obtain. When using 'optgp' this must be a
        multiple of `processes`, otherwise a larger number of samples will be
        returned.
    method : str, optional
        The sampling algorithm to use.
    thinning : int, optional
        The thinning factor of the generated sampling chain. A thinning of 10
        means samples are returned every 10 steps. Defaults to 100 which in
        benchmarks gives approximately uncorrelated samples. If set to one
        will return all iterates.
    processes : int, optional
        Only used for 'optgp'. The number of processes used to generate
        samples.
    seed : int > 0, optional
        The random number seed to be used. Initialized to current time stamp
        if None.

    Returns
    -------
    pandas.DataFrame
        The generated flux samples. Each row corresponds to a sample of the
        fluxes and the columns are the reactions.

    Notes
    -----
    The samplers have a correction method to ensure equality feasibility for
    long-running chains, however this will only work for homogeneous models,
    meaning models with no non-zero fixed variables or constraints (
    right-hand side of the equalities are zero).

    References
    ----------
    .. [1] Megchelenbrink W, Huynen M, Marchiori E (2014)
       optGpSampler: An Improved Tool for Uniformly Sampling the Solution-Space
       of Genome-Scale Metabolic Networks.
       PLoS ONE 9(2): e86587.
    .. [2] Direction Choice for Accelerated Convergence in Hit-and-Run Sampling
       David E. Kaufman Robert L. Smith
       Operations Research 199846:1 , 84-95

    """

    if method == "optgp":
        sampler = OptGPSampler(model, processes, thinning=thinning, seed=seed)
    elif method == "achr":
        sampler = ACHRSampler(model, thinning=thinning, seed=seed)
    else:
        raise ValueError("method must be 'optgp' or 'achr'!")

    return pandas.DataFrame(columns=[rxn.id for rxn in model.reactions],
                            data=sampler.sample(n))



# -*- coding: utf-8 -*-

"""Contains module level fixtures and utility functions."""

from __future__ import absolute_import

from os.path import join
from pickle import dump, load

import pytest


@pytest.fixture(scope="module")
def mini_model(data_directory):
    """Fixture for mini model."""
    with open(join(data_directory, "mini.pickle"), "rb") as infile:
        return load(infile)


def compare_models(model_1, model_2):
    """Compare two models (only for testing purposes)."""
    assert len(model_1.reactions) == len(model_2.reactions)
    assert len(model_1.metabolites) == len(model_2.metabolites)
    assert len(model_1.genes) == len(model_2.genes)
    assert model_1.objective.direction == model_2.objective.direction

    # check Reaction attributes
    for attr in ("id", "name", "lower_bound", "upper_bound",
                 "objective_coefficient", "gene_reaction_rule"):
        assert getattr(model_1.reactions[0], attr) == getattr(
            model_2.reactions[0], attr)
        assert getattr(model_1.reactions[5], attr) == getattr(
            model_2.reactions[5], attr)
        assert getattr(model_1.reactions[-1], attr) == getattr(
            model_2.reactions[-1], attr)

    # check Metabolite attributes
    for attr in ("id", "name", "compartment", "formula", "charge"):
        assert getattr(model_1.metabolites[0], attr) == getattr(
            model_2.metabolites[0], attr)
        assert getattr(model_1.metabolites[5], attr) == getattr(
            model_2.metabolites[5], attr)
        assert getattr(model_1.metabolites[-1], attr) == getattr(
            model_2.metabolites[-1], attr)
        assert len(model_1.reactions[0].metabolites) == len(
            model_2.reactions[0].metabolites)
    # TODO: either relax gene attribute checking or fix models for testing.
    # check Gene attributes
    # for attr in ("id", "name"):
    #     assert getattr(model_1.genes[0], attr) == getattr(model_2.genes[0],
    #                                                       attr)
    #     assert getattr(model_1.genes[10], attr) == getattr(model_2.genes[10],
    #                                                        attr)
    #     assert getattr(model_1.genes[-1], attr) == getattr(model_2.genes[-1],
    #                                                        attr)

    assert len(model_1.reactions[8].metabolites) == len(
        model_2.reactions[8].metabolites)
    assert len(model_1.reactions[-1].metabolites) == len(
        model_2.reactions[-1].metabolites)
    assert len(model_1.genes) == len(model_2.genes)

    # ensure they have the same solution max
    solution_1 = model_1.optimize()
    solution_2 = model_2.optimize()
    assert abs(solution_1.objective_value -
               solution_2.objective_value) == pytest.approx(0.0)

    # ensure the references are correct
    # metabolite -> model reference
    assert model_1.metabolites[0]._model is model_1
    assert model_2.metabolites[0]._model is model_2

    # reaction -> model reference
    assert model_1.reactions[0]._model is model_1
    assert model_2.reactions[0]._model is model_2

    # gene -> model reference
    assert model_1.genes[0]._model is model_1
    assert model_2.genes[0]._model is model_2

    # extra comparisons
    # assert model_1.compartments == model_2.compartments
    # assert dict(model_1.metabolites[4].annotation) == dict(
    #     model_2.metabolites[4].annotation)
    # assert dict(model_1.reactions[4].annotation) == dict(
    #     model_2.reactions[4].annotation)
    # assert dict(model_1.genes[5].annotation) == dict(
    #     model_2.genes[5].annotation)



# -*- coding: utf-8 -*-

"""Test functions of model.py"""

from __future__ import absolute_import

import os
import warnings
from copy import copy, deepcopy
from math import isnan

import numpy as np
import pandas as pd
import pytest
from optlang.symbolics import Zero

import cobra.util.solver as su
from cobra.core import Group, Metabolite, Model, Reaction
from cobra.exceptions import OptimizationError
from cobra.util.solver import SolverNotFound, set_objective, solvers


stable_optlang = ["glpk", "cplex", "gurobi"]
optlang_solvers = ["optlang-" + s for s in stable_optlang if s in su.solvers]


def same_ex(ex1, ex2):
    """Compare to expressions for mathematical equality."""
    return ex1.simplify() == ex2.simplify()


@pytest.mark.parametrize("solver", stable_optlang)
def test_add_remove_reaction_benchmark(model, benchmark, solver):
    metabolite_foo = Metabolite("test_foo")
    metabolite_bar = Metabolite("test_bar")
    metabolite_baz = Metabolite("test_baz")
    actual_metabolite = model.metabolites[0]
    dummy_reaction = Reaction("test_foo_reaction")
    dummy_reaction.add_metabolites({metabolite_foo: -1,
                                    metabolite_bar: 1,
                                    metabolite_baz: -2,
                                    actual_metabolite: 1})

    def benchmark_add_reaction():
        model.add_reaction(dummy_reaction)
        if not getattr(model, 'solver', None):
            solver_dict[solver].create_problem(model)
        model.remove_reactions([dummy_reaction])

    benchmark(benchmark_add_reaction)


def test_add_metabolite(model):
    new_metabolite = Metabolite('test_met')
    assert new_metabolite not in model.metabolites
    with model:
        model.add_metabolites(new_metabolite)
        assert new_metabolite._model == model
        assert new_metabolite in model.metabolites
        assert new_metabolite.id in model.solver.constraints

    assert new_metabolite._model is None
    assert new_metabolite not in model.metabolites
    assert new_metabolite.id not in model.solver.constraints


def test_remove_metabolite_subtractive(model):
    test_metabolite = model.metabolites[4]
    test_reactions = test_metabolite.reactions
    with model:
        model.remove_metabolites(test_metabolite, destructive=False)
        assert test_metabolite._model is None
        assert test_metabolite not in model.metabolites
        assert test_metabolite.id not in model.solver.constraints
        for reaction in test_reactions:
            assert reaction in model.reactions

    assert test_metabolite._model is model
    assert test_metabolite in model.metabolites
    assert test_metabolite.id in model.solver.constraints


def test_remove_metabolite_destructive(model):
    test_metabolite = model.metabolites[4]
    test_reactions = test_metabolite.reactions
    with model:
        model.remove_metabolites(test_metabolite, destructive=True)
        assert test_metabolite._model is None
        assert test_metabolite not in model.metabolites
        assert test_metabolite.id not in model.solver.constraints
        for reaction in test_reactions:
            assert reaction not in model.reactions

    assert test_metabolite._model is model
    assert test_metabolite in model.metabolites
    assert test_metabolite.id in model.solver.constraints
    for reaction in test_reactions:
        assert reaction in model.reactions


def test_compartments(model):
    assert set(model.compartments) == {"c", "e"}
    model = Model("test", "test")
    met_c = Metabolite("a_c", compartment="c")
    met_e = Metabolite("a_e", compartment="e")
    rxn = Reaction("foo")
    rxn.add_metabolites({met_e: -1, met_c: 1})
    model.add_reactions([rxn])
    assert model.compartments == {'c': '', 'e': ''}
    model.compartments = {'c': 'cytosol'}
    assert model.compartments == {'c': 'cytosol', 'e': ''}


def test_add_reaction(model):
    old_reaction_count = len(model.reactions)
    old_metabolite_count = len(model.metabolites)
    dummy_metabolite_1 = Metabolite("test_foo_1")
    dummy_metabolite_2 = Metabolite("test_foo_2")
    actual_metabolite = model.metabolites[0]
    copy_metabolite = model.metabolites[1].copy()
    dummy_reaction = Reaction("test_foo_reaction")
    dummy_reaction.add_metabolites({dummy_metabolite_1: -1,
                                    dummy_metabolite_2: 1,
                                    copy_metabolite: -2,
                                    actual_metabolite: 1})

    model.add_reaction(dummy_reaction)
    assert model.reactions.get_by_id(dummy_reaction.id) == dummy_reaction
    for x in [dummy_metabolite_1, dummy_metabolite_2]:
        assert model.metabolites.get_by_id(x.id) == x
    # Should add 1 reaction and 2 metabolites
    assert len(model.reactions) == old_reaction_count + 1
    assert len(model.metabolites) == old_metabolite_count + 2
    # Test the added reaction
    reaction_in_model = model.reactions.get_by_id(dummy_reaction.id)
    assert type(reaction_in_model) is Reaction
    assert reaction_in_model is dummy_reaction
    assert len(reaction_in_model._metabolites) == 4
    for i in reaction_in_model._metabolites:
        assert type(i) == Metabolite
    # Test the added metabolites
    met1_in_model = model.metabolites.get_by_id(dummy_metabolite_1.id)
    assert met1_in_model is dummy_metabolite_1
    copy_in_model = model.metabolites.get_by_id(copy_metabolite.id)
    assert copy_metabolite is not copy_in_model
    assert type(copy_in_model) is Metabolite
    assert dummy_reaction in actual_metabolite._reaction
    # Test addition of a different metabolite with the same name as an
    # existing one uses the metabolite in the model
    r2 = Reaction("test_foo_reaction2")
    model.add_reaction(r2)
    r2.add_metabolites({Metabolite(model.metabolites[0].id): 1})
    assert model.metabolites[0] is list(r2._metabolites)[0]


def test_add_reaction_context(model):
    old_reaction_count = len(model.reactions)
    old_metabolite_count = len(model.metabolites)
    dummy_metabolite_1 = Metabolite("test_foo_1")
    dummy_metabolite_2 = Metabolite("test_foo_2")
    actual_metabolite = model.metabolites[0]
    copy_metabolite = model.metabolites[1].copy()
    dummy_reaction = Reaction("test_foo_reaction")
    dummy_reaction.add_metabolites({dummy_metabolite_1: -1,
                                    dummy_metabolite_2: 1,
                                    copy_metabolite: -2,
                                    actual_metabolite: 1})
    dummy_reaction.gene_reaction_rule = 'dummy_gene'

    with model:
        model.add_reaction(dummy_reaction)
        assert model.reactions.get_by_id(
            dummy_reaction.id) == dummy_reaction
        assert len(model.reactions) == old_reaction_count + 1
        assert len(model.metabolites) == old_metabolite_count + 2
        assert dummy_metabolite_1._model == model
        assert 'dummy_gene' in model.genes

    assert len(model.reactions) == old_reaction_count
    assert len(model.metabolites) == old_metabolite_count
    with pytest.raises(KeyError):
        model.reactions.get_by_id(dummy_reaction.id)
    assert dummy_metabolite_1._model is None
    assert 'dummy_gene' not in model.genes


def test_add_reaction_from_other_model(model):
    other = model.copy()
    for i in other.reactions:
        i.id += "_other"
    other.repair()
    model.add_reactions(other.reactions)

    # Check if the other reaction has an error in its GPR
    m1 = model.copy()
    m2 = model.copy()
    m1.reactions.PGI.remove_from_model()
    m2.genes.b4025._reaction.clear()
    m1.add_reaction(m2.reactions.PGI)


def test_model_remove_reaction(model):
    old_reaction_count = len(model.reactions)

    with model:
        model.remove_reactions(["PGI"])
        assert len(model.reactions) == old_reaction_count - 1
        with pytest.raises(KeyError):
            model.reactions.get_by_id("PGI")
        model.remove_reactions(model.reactions[:1])
        assert len(model.reactions) == old_reaction_count - 2

    assert len(model.reactions) == old_reaction_count
    assert "PGI" in model.reactions

    tmp_metabolite = Metabolite("testing")
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    assert tmp_metabolite in model.metabolites
    model.remove_reactions(model.reactions[:1],
                           remove_orphans=True)
    assert tmp_metabolite not in model.metabolites

    with model:
        model.reactions[0].add_metabolites({tmp_metabolite: 1})
        assert tmp_metabolite in model.metabolites
    assert tmp_metabolite not in model.metabolites

    biomass_before = model.slim_optimize()
    with model:
        model.remove_reactions([model.reactions.Biomass_Ecoli_core])
        assert np.isclose(model.slim_optimize(), 0)

    assert np.isclose(model.slim_optimize(), biomass_before)


def test_reaction_remove(model):
    old_reaction_count = len(model.reactions)
    tmp_metabolite = Metabolite("testing")

    # Delete without removing orphan
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    assert len(tmp_metabolite.reactions) == 1

    # Esnsure the stoichiometry is still the same using different objects
    removed_reaction = model.reactions[0]
    original_stoich = {i.id: value for i, value
                       in removed_reaction._metabolites.items()}
    model.reactions[0].remove_from_model(remove_orphans=False)
    assert len(original_stoich) == len(removed_reaction._metabolites)
    for met in removed_reaction._metabolites:
        assert original_stoich[met.id] == removed_reaction._metabolites[
            met]
        assert met is not model.metabolites

    # Make sure it's still in the model
    assert tmp_metabolite in model.metabolites
    assert len(tmp_metabolite.reactions) == 0
    assert len(model.reactions) == old_reaction_count - 1

    # Now try with removing orphans
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    assert len(tmp_metabolite.reactions) == 1
    model.reactions[0].remove_from_model(remove_orphans=True)
    assert tmp_metabolite not in model.metabolites
    assert len(tmp_metabolite.reactions) == 0
    assert len(model.reactions) == old_reaction_count - 2

    # It shouldn't remove orphans if it's in 2 reactions however
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    model.reactions[1].add_metabolites({tmp_metabolite: 1})
    assert len(tmp_metabolite.reactions) == 2
    model.reactions[0].remove_from_model(remove_orphans=False)
    assert tmp_metabolite in model.metabolites
    assert len(tmp_metabolite.reactions) == 1
    assert len(model.reactions) == old_reaction_count - 3


def test_reaction_delete(model):
    old_reaction_count = len(model.reactions)
    tmp_metabolite = Metabolite("testing")

    # Delete without removing orphan
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    assert len(tmp_metabolite.reactions) == 1
    with pytest.warns(DeprecationWarning):
        model.reactions[0].delete(remove_orphans=False)

    # Make sure it's still in the model
    assert tmp_metabolite in model.metabolites
    assert len(tmp_metabolite.reactions) == 0
    assert len(model.reactions) == old_reaction_count - 1

    # Now try it with removing orphans
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    assert len(tmp_metabolite.reactions) == 1
    model.reactions[0].delete(remove_orphans=True)
    assert tmp_metabolite not in model.metabolites
    assert len(tmp_metabolite.reactions) == 0
    assert len(model.reactions) == old_reaction_count - 2

    # It shouldn't remove orphans if it's in 2 reactions however
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    model.reactions[1].add_metabolites({tmp_metabolite: 1})
    assert len(tmp_metabolite.reactions) == 2
    model.reactions[0].delete(remove_orphans=False)
    assert tmp_metabolite in model.metabolites
    assert len(tmp_metabolite.reactions) == 1
    assert len(model.reactions) == old_reaction_count - 3


def test_remove_gene(model):
    target_gene = model.genes[0]
    gene_reactions = list(target_gene.reactions)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        target_gene.remove_from_model()
    assert target_gene.model is None

    # Make sure the reaction was removed from the model
    assert target_gene not in model.genes

    # Ensure the old reactions no longer have a record of the gene
    for reaction in gene_reactions:
        assert target_gene not in reaction.genes


def test_group_model_reaction_association(model):
    num_members = 5
    reactions_for_group = model.reactions[0:num_members]
    group = Group("arbitrary_group1")
    group.add_members(reactions_for_group)
    group.kind = "collection"
    model.add_groups([group])
    # group should point to and be associated with the model
    assert group._model is model
    assert group in model.groups

    # model.get_associated_groups should find the group for each reaction
    # we added to the group
    for reaction in reactions_for_group:
        assert group in model.get_associated_groups(reaction)

    # remove the group from the model and check that reactions are no
    # longer associated with the group
    model.remove_groups([group])
    assert group not in model.groups
    assert group._model is not model
    for reaction in reactions_for_group:
        assert group not in model.get_associated_groups(reaction)


def test_group_members_add_to_model(model):
    # remove a few reactions from the model and add them to a new group
    num_members = 5
    reactions_for_group = model.reactions[0:num_members]
    model.remove_reactions(reactions_for_group, remove_orphans=False)
    group = Group("arbitrary_group1")
    group.add_members(reactions_for_group)
    group.kind = "collection"
    # the old reactions should not be in the model
    for reaction in reactions_for_group:
        assert reaction not in model.reactions

    # add the group to the model and check that the reactions were added
    model.add_groups([group])
    assert group in model.groups
    for reaction in reactions_for_group:
        assert reaction in model.reactions


def test_group_loss_of_elements(model):
    # when a metabolite, reaction, or gene is removed from a model, it
    # should no longer be a member of any groups
    num_members_each = 5
    elements_for_group = model.reactions[0:num_members_each]
    elements_for_group.extend(model.metabolites[0:num_members_each])
    elements_for_group.extend(model.genes[0:num_members_each])
    group = Group("arbitrary_group1")
    group.add_members(elements_for_group)
    group.kind = "collection"
    model.add_groups([group])

    remove_met = model.metabolites[0]
    model.remove_metabolites([remove_met])
    remove_rxn = model.reactions[0]
    model.remove_reactions([remove_rxn])
    remove_gene = model.genes[0]
    remove_gene.remove_from_model()
    assert remove_met not in group.members
    assert remove_rxn not in group.members
    assert remove_gene not in group.members


def test_exchange_reactions(model):
    assert set(model.exchanges) == set([rxn for rxn in model.reactions
                                        if rxn.id.startswith("EX")])


@pytest.mark.parametrize("metabolites, reaction_type, prefix", [
    ("exchange", "exchange", "EX_"),
    ("demand", "demand", "DM_"),
    ("sink", "sink", "SK_")
], indirect=["metabolites"])
def test_add_boundary(model, metabolites, reaction_type, prefix):
    for metabolite in metabolites:
        reaction = model.add_boundary(metabolite, reaction_type)
        assert model.reactions.get_by_id(
            reaction.id) == reaction
        assert reaction.reactants == [metabolite]
        assert model.constraints[metabolite.id].expression.has(
            model.variables[prefix + metabolite.id])


@pytest.mark.parametrize("metabolites, reaction_type, prefix", [
    ("exchange", "exchange", "EX_"),
    ("demand", "demand", "DM_"),
    ("sink", "sink", "SK_")
], indirect=["metabolites"])
def test_add_boundary_context(model, metabolites, reaction_type,
                              prefix):
    with model:
        for metabolite in metabolites:
            reaction = model.add_boundary(metabolite, reaction_type)
            assert model.reactions.get_by_id(
                reaction.id) == reaction
            assert reaction.reactants == [metabolite]
            assert -model.constraints[
                metabolite.id].expression.has(
                model.variables[prefix + metabolite.id])
    for metabolite in metabolites:
        assert prefix + metabolite.id not in model.reactions
        assert prefix + metabolite.id not in model.variables.keys()


@pytest.mark.parametrize("metabolites, reaction_type", [
    ("exchange", "exchange"),
    ("demand", "demand"),
    ("sink", "sink")
], indirect=["metabolites"])
def test_add_existing_boundary(model, metabolites, reaction_type):
    for metabolite in metabolites:
        model.add_boundary(metabolite, reaction_type)
        with pytest.raises(ValueError):
            model.add_boundary(metabolite, reaction_type)


@pytest.mark.parametrize("solver", stable_optlang)
def test_copy_benchmark(model, solver, benchmark):
    def _():
        model.copy()
        if not getattr(model, 'solver', None):
            solver_dict[solver].create_problem(model)

    benchmark(_)


@pytest.mark.parametrize("solver", stable_optlang)
def test_copy_benchmark_large_model(large_model, solver, benchmark):
    def _():
        large_model.copy()
        if not getattr(large_model, 'solver', None):
            solver_dict[solver].create_problem(large_model)

    benchmark(_)


def test_copy(model):
    # Modifying copy should not modify the original
    # Test that deleting reactions in the copy does not change the
    # number of reactions in the original model
    model_copy = model.copy()
    old_reaction_count = len(model.reactions)
    assert model_copy.notes is not model.notes
    assert model_copy.annotation is not model.annotation
    assert len(model.reactions) == len(model_copy.reactions)
    assert len(model.metabolites) == len(model_copy.metabolites)
    model_copy.remove_reactions(model_copy.reactions[0:5])
    assert old_reaction_count == len(model.reactions)
    assert len(model.reactions) != len(model_copy.reactions)
    # Copying a model should not copy its context
    with model:
        model.remove_reactions([model.reactions.ACALD])
        cp_model = model.copy()
        assert len(cp_model._contexts) == 0
    assert 'ACALD' not in cp_model.reactions


def test_copy_with_groups(model):
    sub = Group("pathway", members=[model.reactions.PFK, model.reactions.FBA])
    model.add_groups([sub])
    copy = model.copy()
    assert len(copy.groups) == len(model.groups)
    assert len(copy.groups.get_by_id("pathway")) == len(
        model.groups.get_by_id("pathway"))


def test_deepcopy_benchmark(model, benchmark):
    benchmark(deepcopy, model)


def test_deepcopy(model):
    # Reference structures are maintained when deepcopying
    model_copy = deepcopy(model)
    for gene, gene_copy in zip(model.genes, model_copy.genes):
        assert gene.id == gene_copy.id
        reactions = sorted(i.id for i in gene.reactions)
        reactions_copy = sorted(i.id for i in gene_copy.reactions)
        assert reactions == reactions_copy
    for reaction, reaction_copy in zip(model.reactions,
                                       model_copy.reactions):
        assert reaction.id == reaction_copy.id
        metabolites = sorted(i.id for i in reaction._metabolites)
        metabolites_copy = sorted(i.id for i in reaction_copy._metabolites)
        assert metabolites == metabolites_copy


def test_add_reaction_orphans(model):
    # Test reaction addition
    # Need to verify that no orphan genes or metabolites are
    # contained in reactions after adding them to the model.
    model = model.__class__('test')
    model.add_reactions((x.copy() for x in model.reactions))
    genes = []
    metabolites = []
    for x in model.reactions:
        genes.extend(x.genes)
        metabolites.extend(x._metabolites)
    orphan_genes = [x for x in genes if x.model is not model]
    orphan_metabolites = [x for x in metabolites if x.model is not model]
    # Check for dangling genes when running Model.add_reactions
    assert len(orphan_genes) == 0
    # Check for dangling metabolites when running Model.add_reactions
    assert len(orphan_metabolites) == 0


def test_merge_models(model, tiny_toy_model):
    with model, tiny_toy_model:
        # Add some cons/vars to tiny_toy_model for testing merging
        tiny_toy_model.add_reactions([Reaction('EX_glc__D_e')])
        variable = tiny_toy_model.problem.Variable('foo')
        constraint = tiny_toy_model.problem.Constraint(
            variable, ub=0, lb=0, name='constraint')
        tiny_toy_model.add_cons_vars([variable, constraint])

        # Test merging to new model
        merged = model.merge(tiny_toy_model, inplace=False,
                             objective='sum', prefix_existing='tiny_')
        assert 'ex1' in merged.reactions
        assert 'ex1' not in model.reactions
        assert merged.reactions.ex1.objective_coefficient == 1
        assert merged.reactions.get_by_id(
            'Biomass_Ecoli_core').objective_coefficient == 1
        assert 'tiny_EX_glc__D_e' in merged.reactions
        assert 'foo' in merged.variables

        # Test reversible in-place model merging
        with model:
            model.merge(tiny_toy_model, inplace=True, objective='left',
                        prefix_existing='tiny_')
            assert 'ex1' in model.reactions
            assert 'constraint' in model.constraints
            assert 'foo' in model.variables
            assert 'tiny_EX_glc__D_e' in model.reactions
            assert (model.objective.expression.simplify() ==
                    model.reactions.get_by_id(
                        'Biomass_Ecoli_core').flux_expression.simplify())
        assert 'ex1' not in model.reactions
        assert 'constraint' not in model.constraints
        assert 'foo' not in model.variables
        assert 'tiny_EX_glc__D_e' not in model.reactions

    # Test the deprecated operator overloading
    with model:
        merged = model + tiny_toy_model
        assert 'ex1' in merged.reactions
    with model:
        model += tiny_toy_model
        assert 'ex1' in model.reactions


@pytest.mark.parametrize("solver", stable_optlang)
def test_change_objective_benchmark(model, benchmark, solver):
    atpm = model.reactions.get_by_id("ATPM")

    def benchmark_change_objective():
        model.objective = atpm.id
        if not getattr(model, 'solver', None):
            solver_dict[solver].create_problem(model)

    benchmark(benchmark_change_objective)


def test_get_objective_direction(model):
    assert model.objective_direction == "max"
    value = model.slim_optimize()
    assert np.isclose(value, 0.874, 1e-3)


def test_set_objective_direction(model):
    with model:
        model.objective_direction = "min"
        assert model.objective_direction == "min"
        value = model.slim_optimize()
        assert value == 0.0
    assert model.objective_direction == "max"


def test_slim_optimize(model):
    with model:
        assert model.slim_optimize() > 0.872
        model.reactions.Biomass_Ecoli_core.lower_bound = 10
        assert isnan(model.slim_optimize())
        with pytest.raises(OptimizationError):
            model.slim_optimize(error_value=None)


@pytest.mark.parametrize("solver", optlang_solvers)
def test_optimize(model, solver):
    model.solver = solver
    with model:
        assert model.optimize().objective_value > 0.872
        model.reactions.Biomass_Ecoli_core.lower_bound = 10
        with pytest.warns(UserWarning):
            model.optimize()
        with pytest.raises(OptimizationError):
            model.optimize(raise_error=True)


def test_change_objective(model):
    # Test for correct optimization behavior
    model.optimize()
    assert model.reactions.Biomass_Ecoli_core.x > 0.5
    with model:
        model.objective = model.reactions.EX_etoh_e
        model.optimize()
    assert model.reactions.Biomass_Ecoli_core.x < 0.5
    assert model.reactions.Biomass_Ecoli_core.objective_coefficient == 1
    model.optimize()
    assert model.reactions.Biomass_Ecoli_core.x > 0.5
    # Test changing objective
    biomass = model.reactions.get_by_id("Biomass_Ecoli_core")
    atpm = model.reactions.get_by_id("ATPM")
    model.objective = atpm.id
    assert atpm.objective_coefficient == 1.
    assert biomass.objective_coefficient == 0.
    assert su.linear_reaction_coefficients(model) == {atpm: 1.}
    # Change it back using object itself
    model.objective = biomass
    assert atpm.objective_coefficient == 0.
    assert biomass.objective_coefficient == 1.
    # Set both to 1 with a list
    model.objective = [atpm, biomass]
    assert atpm.objective_coefficient == 1.
    assert biomass.objective_coefficient == 1.
    # Set both using a dict
    model.objective = {atpm: 0.2, biomass: 0.3}
    assert abs(atpm.objective_coefficient - 0.2) < 10 ** -9
    assert abs(biomass.objective_coefficient - 0.3) < 10 ** -9
    # Test setting by index
    model.objective = model.reactions.index(atpm)
    assert su.linear_reaction_coefficients(model) == {atpm: 1.}
    # Test by setting list of indexes
    model.objective = [model.reactions.index(reaction) for
                       reaction in [atpm, biomass]]
    assert su.linear_reaction_coefficients(model) == {atpm: 1.,
                                                      biomass: 1.}


def test_problem_properties(model):
    new_variable = model.problem.Variable("test_variable")
    new_constraint = model.problem.Constraint(Zero,
                                              name="test_constraint")
    model.add_cons_vars([new_variable, new_constraint])
    assert "test_variable" in model.variables.keys()
    assert "test_constraint" in model.constraints.keys()
    model.remove_cons_vars([new_constraint, new_variable])
    assert "test_variable" not in model.variables.keys()
    assert "test_constraint" not in model.variables.keys()


def test_solution_data_frame(model):
    solution = model.optimize().to_frame()
    assert isinstance(solution, pd.DataFrame)
    assert 'fluxes' in solution
    assert 'reduced_costs' in solution


def test_context_manager(model):
    bounds0 = model.reactions[0].bounds
    bounds1 = (1, 2)
    bounds2 = (3, 4)

    # Trigger a nested model context, ensuring that bounds are
    # preserved at each level
    with model:
        model.reactions[0].bounds = bounds1
        with model:
            model.reactions[0].bounds = bounds2

            assert model.reactions[0].bounds == bounds2
        assert model.reactions[0].bounds == bounds1
    assert model.reactions[0].bounds == bounds0


def test_objective_coefficient_reflects_changed_objective(model):
    biomass_r = model.reactions.get_by_id('Biomass_Ecoli_core')
    assert biomass_r.objective_coefficient == 1
    model.objective = "PGI"
    assert biomass_r.objective_coefficient == 0
    assert model.reactions.PGI.objective_coefficient == 1


def test_change_objective_through_objective_coefficient(model):
    biomass_r = model.reactions.get_by_id('Biomass_Ecoli_core')
    pgi = model.reactions.PGI
    pgi.objective_coefficient = 2
    coef_dict = model.objective.expression.as_coefficients_dict()
    # Check that objective has been updated
    assert coef_dict[pgi.forward_variable] == 2.0
    assert coef_dict[pgi.reverse_variable] == -2.0
    # Check that original objective is still in there
    assert coef_dict[biomass_r.forward_variable] == 1.0
    assert coef_dict[biomass_r.reverse_variable] == -1.0


def test_transfer_objective(model):
    new_mod = Model("new model")
    new_mod.add_reactions(model.reactions)
    new_mod.objective = model.objective
    assert (set(str(x) for x in model.objective.expression.args) == set(
        str(x) for x in new_mod.objective.expression.args))
    new_mod.slim_optimize()
    assert abs(new_mod.objective.value - 0.874) < 0.001


def test_model_from_other_model(model):
    model = Model(id_or_model=model)
    for reaction in model.reactions:
        assert reaction == model.reactions.get_by_id(reaction.id)


def test_add_reactions(model):
    r1 = Reaction('r1')
    r1.add_metabolites({Metabolite('A'): -1, Metabolite('B'): 1})
    r1.lower_bound, r1.upper_bound = -999999., 999999.
    r2 = Reaction('r2')
    r2.add_metabolites(
        {Metabolite('A'): -1, Metabolite('C'): 1, Metabolite('D'): 1})
    r2.lower_bound, r2.upper_bound = 0., 999999.
    model.add_reactions([r1, r2])
    r2.objective_coefficient = 3.
    assert r2.objective_coefficient == 3.
    assert model.reactions[-2] == r1
    assert model.reactions[-1] == r2
    assert isinstance(model.reactions[-2].reverse_variable,
                      model.problem.Variable)
    coefficients_dict = model.objective.expression. \
        as_coefficients_dict()
    biomass_r = model.reactions.get_by_id('Biomass_Ecoli_core')
    assert coefficients_dict[biomass_r.forward_variable] == 1.
    assert coefficients_dict[biomass_r.reverse_variable] == -1.
    assert coefficients_dict[
               model.reactions.r2.forward_variable] == 3.
    assert coefficients_dict[
               model.reactions.r2.reverse_variable] == -3.


def test_add_reactions_single_existing(model):
    rxn = model.reactions[0]
    r1 = Reaction(rxn.id)
    r1.add_metabolites({Metabolite('A'): -1, Metabolite('B'): 1})
    r1.lower_bound, r1.upper_bound = -999999., 999999.
    model.add_reactions([r1])
    assert rxn in model.reactions
    assert r1 is not model.reactions.get_by_id(rxn.id)


def test_add_reactions_duplicate(model):
    rxn = model.reactions[0]
    r1 = Reaction('r1')
    r1.add_metabolites({Metabolite('A'): -1, Metabolite('B'): 1})
    r1.lower_bound, r1.upper_bound = -999999., 999999.
    r2 = Reaction(rxn.id)
    r2.add_metabolites(
        {Metabolite('A'): -1, Metabolite('C'): 1, Metabolite('D'): 1})
    model.add_reactions([r1, r2])
    assert r1 in model.reactions
    assert rxn in model.reactions
    assert r2 is not model.reactions.get_by_id(rxn.id)


def test_add_cobra_reaction(model):
    r = Reaction(id="c1")
    model.add_reaction(r)
    assert isinstance(model.reactions.c1, Reaction)


def test_all_objects_point_to_all_other_correct_objects(model):
    for reaction in model.reactions:
        assert reaction.model == model
        for gene in reaction.genes:
            assert gene == model.genes.get_by_id(gene.id)
            assert gene.model == model
            for reaction2 in gene.reactions:
                assert reaction2.model == model
                assert reaction2 == model.reactions.get_by_id(
                    reaction2.id)

        for metabolite in reaction.metabolites:
            assert metabolite.model == model
            assert metabolite == model.metabolites.get_by_id(
                metabolite.id)
            for reaction2 in metabolite.reactions:
                assert reaction2.model == model
                assert reaction2 == model.reactions.get_by_id(
                    reaction2.id)


def test_objects_point_to_correct_other_after_copy(model):
    for reaction in model.reactions:
        assert reaction.model == model
        for gene in reaction.genes:
            assert gene == model.genes.get_by_id(gene.id)
            assert gene.model == model
            for reaction2 in gene.reactions:
                assert reaction2.model == model
                assert reaction2 == model.reactions.get_by_id(
                    reaction2.id)

        for metabolite in reaction.metabolites:
            assert metabolite.model == model
            assert metabolite == model.metabolites.get_by_id(
                metabolite.id)
            for reaction2 in metabolite.reactions:
                assert reaction2.model == model
                assert reaction2 == model.reactions.get_by_id(
                    reaction2.id)


def test_remove_reactions(model):
    reactions_to_remove = model.reactions[10:30]
    assert all([reaction.model is model for reaction in
                reactions_to_remove])
    assert all(
        [model.reactions.get_by_id(reaction.id) == reaction for
         reaction in reactions_to_remove])

    model.remove_reactions(reactions_to_remove)
    assert all(
        [reaction.model is None for reaction in reactions_to_remove])
    for reaction in reactions_to_remove:
        assert reaction.id not in list(
            model.variables.keys())

    model.add_reactions(reactions_to_remove)
    for reaction in reactions_to_remove:
        assert reaction in model.reactions


def test_objective(model):
    obj = model.objective
    assert obj.get_linear_coefficients(obj.variables) == {
        model.variables["Biomass_Ecoli_core_reverse_2cdba"]: -1,
        model.variables["Biomass_Ecoli_core"]: 1
    }
    assert obj.direction == "max"


def test_change_objective(model):
    expression = 1.0 * model.variables['ENO'] + \
                 1.0 * model.variables['PFK']
    model.objective = model.problem.Objective(
        expression)
    assert same_ex(model.objective.expression, expression)
    model.objective = "ENO"
    eno_obj = model.problem.Objective(
        model.reactions.ENO.flux_expression, direction="max")
    pfk_obj = model.problem.Objective(
        model.reactions.PFK.flux_expression, direction="max")
    assert same_ex(model.objective.expression, eno_obj.expression)

    with model:
        model.objective = "PFK"
        assert same_ex(model.objective.expression, pfk_obj.expression)
    assert same_ex(model.objective.expression, eno_obj.expression)
    expression = model.objective.expression
    atpm = model.reactions.get_by_id("ATPM")
    biomass = model.reactions.get_by_id("Biomass_Ecoli_core")
    with model:
        model.objective = atpm
    assert same_ex(model.objective.expression, expression)
    with model:
        atpm.objective_coefficient = 1
        biomass.objective_coefficient = 2
    assert same_ex(model.objective.expression, expression)

    with model:
        set_objective(model, model.problem.Objective(
            atpm.flux_expression))
        assert same_ex(model.objective.expression, atpm.flux_expression)
    assert same_ex(model.objective.expression, expression)

    expression = model.objective.expression
    with model:
        with model:  # Test to make sure nested contexts are OK
            set_objective(model, atpm.flux_expression,
                          additive=True)
            assert same_ex(model.objective.expression,
                           expression + atpm.flux_expression)
    assert same_ex(model.objective.expression, expression)


def test_set_reaction_objective(model):
    model.objective = model.reactions.ACALD
    assert str(model.objective.expression) == str(
        1.0 * model.reactions.ACALD.forward_variable -
        1.0 * model.reactions.ACALD.reverse_variable)


def test_set_reaction_objective_str(model):
    model.objective = model.reactions.ACALD.id
    assert str(model.objective.expression) == str(
        1.0 * model.reactions.ACALD.forward_variable -
        1.0 * model.reactions.ACALD.reverse_variable)


def test_invalid_objective_raises(model):
    with pytest.raises(ValueError):
        setattr(model, 'objective', 'This is not a valid objective!')
    with pytest.raises(TypeError):
        setattr(model, 'objective', 3.)


@pytest.mark.skipif("cplex" not in solvers, reason="need cplex")
def test_solver_change(model):
    solver_id = id(model.solver)
    problem_id = id(model.solver.problem)
    solution = model.optimize().fluxes
    model.solver = "cplex"
    assert id(model.solver) != solver_id
    assert id(model.solver.problem) != problem_id
    new_solution = model.optimize().fluxes
    assert np.allclose(solution, new_solution, rtol=0, atol=1E-06)


def test_no_change_for_same_solver(model):
    solver_id = id(model.solver)
    problem_id = id(model.solver.problem)
    model.solver = "glpk"
    assert id(model.solver) == solver_id
    assert id(model.solver.problem) == problem_id


def test_invalid_solver_change_raises(model):
    with pytest.raises(SolverNotFound):
        setattr(model, 'solver', [1, 2, 3])
    with pytest.raises(SolverNotFound):
        setattr(model, 'solver',
                'ThisIsDefinitelyNotAvalidSolver')
    with pytest.raises(SolverNotFound):
        setattr(model, 'solver', os)


@pytest.mark.skipif('cplex' not in solvers, reason='no cplex')
def test_change_solver_to_cplex_and_check_copy_works(model):
    assert (model.slim_optimize() - 0.8739215069684306) == pytest.approx(0.0)
    model_copy = model.copy()
    assert (model_copy.slim_optimize() - 0.8739215069684306) == pytest.approx(
        0.0)
    # Second, change existing glpk based model to cplex
    model.solver = 'cplex'
    assert (model.slim_optimize() - 0.8739215069684306) == pytest.approx(0.0)
    model_copy = copy(model)
    assert (model_copy.slim_optimize() - 0.8739215069684306) == pytest.approx(
        0.0)


def test_copy_preserves_existing_solution(solved_model):
    solution, model = solved_model
    model_cp = copy(model)
    primals_original = [variable.primal for variable in
                        model.variables]
    primals_copy = [variable.primal for variable in
                    model_cp.variables]
    abs_diff = abs(
        np.array(primals_copy) - np.array(primals_original))
    assert not any(abs_diff > 1e-6)


def test_repr_html_(model):
    assert '<table>' in model._repr_html_()



# -*- coding: utf-8 -*-

"""Module level fixtures"""

from __future__ import absolute_import

import pytest

from cobra.util.solver import solvers


solver_trials = ['glpk',
                 pytest.param('cplex',
                              marks=pytest.mark.skipif(
                                  'cplex' not in solvers,
                                  reason='No CPLEX found on PYTHONPATH')),
                 pytest.param('gurobi',
                              marks=pytest.mark.skipif(
                                  'gurobi' not in solvers,
                                  reason='No Gurobi found on PYTHONPATH'))]


@pytest.fixture(scope="function", params=solver_trials)
def solved_model(request, model):
    model.solver = request.param
    solution = model.optimize()
    return solution, model



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from cobra.core.summary.summary import Summary
from cobra.core.summary.metabolite_summary import MetaboliteSummary
from cobra.core.summary.model_summary import ModelSummary
from cobra.core.summary.reaction_summary import ReactionSummary



# -*- coding: utf-8 -*-

from __future__ import absolute_import

from ast import NodeTransformer
from itertools import chain
from warnings import warn

from six import iteritems

from cobra.core import Reaction
from cobra.core.gene import ast2str
from cobra.manipulation.delete import get_compiled_gene_reaction_rules
from cobra.util.solver import set_objective


_renames = (
    (".", "_DOT_"),
    ("(", "_LPAREN_"),
    (")", "_RPAREN_"),
    ("-", "__"),
    ("[", "_LSQBKT"),
    ("]", "_RSQBKT"),
    (",", "_COMMA_"),
    (":", "_COLON_"),
    (">", "_GT_"),
    ("<", "_LT"),
    ("/", "_FLASH"),
    ("\\", "_BSLASH"),
    ("+", "_PLUS_"),
    ("=", "_EQ_"),
    (" ", "_SPACE_"),
    ("'", "_SQUOT_"),
    ('"', "_DQUOT_"),
)


def _escape_str_id(id_str):
    """make a single string id SBML compliant"""
    for c in ("'", '"'):
        if id_str.startswith(c) and id_str.endswith(c) \
                and id_str.count(c) == 2:
            id_str = id_str.strip(c)
    for char, escaped_char in _renames:
        id_str = id_str.replace(char, escaped_char)
    return id_str


class _GeneEscaper(NodeTransformer):
    def visit_Name(self, node):
        node.id = _escape_str_id(node.id)
        return node


def escape_ID(cobra_model):
    """makes all ids SBML compliant"""
    for x in chain([cobra_model],
                   cobra_model.metabolites,
                   cobra_model.reactions,
                   cobra_model.genes):
        x.id = _escape_str_id(x.id)
    cobra_model.repair()
    gene_renamer = _GeneEscaper()
    for rxn, rule in iteritems(get_compiled_gene_reaction_rules(cobra_model)):
        if rule is not None:
            rxn._gene_reaction_rule = ast2str(gene_renamer.visit(rule))


def rename_genes(cobra_model, rename_dict):
    """renames genes in a model from the rename_dict"""
    recompute_reactions = set()  # need to recomptue related genes
    remove_genes = []
    for old_name, new_name in iteritems(rename_dict):
        # undefined if there a value matches a different key
        # because dict is unordered
        try:
            gene_index = cobra_model.genes.index(old_name)
        except ValueError:
            gene_index = None
        old_gene_present = gene_index is not None
        new_gene_present = new_name in cobra_model.genes
        if old_gene_present and new_gene_present:
            old_gene = cobra_model.genes.get_by_id(old_name)
            # Added in case not renaming some genes:
            if old_gene is not cobra_model.genes.get_by_id(new_name):
                remove_genes.append(old_gene)
                recompute_reactions.update(old_gene._reaction)
        elif old_gene_present and not new_gene_present:
            # rename old gene to new gene
            gene = cobra_model.genes[gene_index]
            # trick DictList into updating index
            cobra_model.genes._dict.pop(gene.id)  # ugh
            gene.id = new_name
            cobra_model.genes[gene_index] = gene
        elif not old_gene_present and new_gene_present:
            pass
        else:  # if not old gene_present and not new_gene_present
            # the new gene's _model will be set by repair
            # This would add genes from rename_dict
            # that are not associated with a rxn
            # cobra_model.genes.append(Gene(new_name))
            pass
    cobra_model.repair()

    class Renamer(NodeTransformer):
        def visit_Name(self, node):
            node.id = rename_dict.get(node.id, node.id)
            return node

    gene_renamer = Renamer()
    for rxn, rule in iteritems(get_compiled_gene_reaction_rules(cobra_model)):
        if rule is not None:
            rxn._gene_reaction_rule = ast2str(gene_renamer.visit(rule))

    for rxn in recompute_reactions:
        rxn.gene_reaction_rule = rxn._gene_reaction_rule
    for i in remove_genes:
        cobra_model.genes.remove(i)



# -*- coding: utf-8 -*-

"""Test functionalities of ModelSummary."""

from __future__ import absolute_import

import numpy as np
import pytest

from cobra.test.test_core.test_summary import (
    captured_output, check_in_line, check_line)


@pytest.mark.parametrize("names", [False, True])
def test_model_summary_to_table_previous_solution(model, opt_solver, names):
    """Test Summary._to_table() of previous solution."""
    model.solver = opt_solver
    solution = model.optimize()
    rxn_test = model.exchanges[0]

    if names:
        met_test = list(rxn_test.metabolites.keys())[0].name
    else:
        met_test = list(rxn_test.metabolites.keys())[0].id

    solution.fluxes[rxn_test.id] = 321

    with captured_output() as (out, _):
        print(model.summary(solution, names=names))
        check_in_line(out.getvalue(), [met_test + '321'])


@pytest.mark.parametrize("names", [False, True])
def test_model_summary_to_frame_previous_solution(model, opt_solver, names):
    """Test Summary.to_frame() of previous solution."""
    model.solver = opt_solver
    solution = model.optimize()
    rxn_test = model.exchanges[0]
    solution.fluxes[rxn_test.id] = 321
    out_df = model.summary(solution, names=names).to_frame()

    assert out_df.loc[0, ('OUT_FLUXES', 'FLUX')] == 321


@pytest.mark.parametrize("names", [False, True])
def test_model_summary_to_table(model, opt_solver, names):
    """Test model.summary()._to_table()."""
    model.solver = opt_solver
    # test non-fva version (these should be fixed for textbook model)
    if names:
        expected_entry = ['        O2    21.8       H2O       29.2     '
                          'Biomass Objective Function with GAM   0.874   ']
    else:
        expected_entry = ['     o2_e    21.8      h2o_e      29.2     '
                          'Biomass_Ecoli_core   0.874   ']

    model.optimize()

    with captured_output() as (out, _):
        print(model.summary(names=names))
        check_in_line(out.getvalue(), expected_entry)


@pytest.mark.parametrize("names", [False, True])
def test_model_summary_to_frame(model, opt_solver, names):
    """Test model.summary().to_frame()."""
    model.solver = opt_solver
    # test non-fva version (these should be fixed for textbook model)
    if names:
        expected_in_fluxes = ['O2', 'D-Glucose', 'Ammonium', 'Phosphate']
        expected_out_fluxes = ['H2O', 'CO2', 'H+', np.nan]
    else:
        expected_in_fluxes = ['o2_e', 'glc__D_e', 'nh4_e', 'pi_e']
        expected_out_fluxes = ['h2o_e', 'co2_e', 'h_e', np.nan]

    model.optimize()

    out_df = model.summary(names=names).to_frame()

    assert out_df[('IN_FLUXES', 'ID')].tolist() == expected_in_fluxes
    assert out_df[('OUT_FLUXES', 'ID')].tolist() == expected_out_fluxes


@pytest.mark.parametrize("fraction", [0.95])
def test_model_summary_to_table_with_fva(model, opt_solver, fraction):
    """Test model summary._to_table() (using FVA)."""
    if opt_solver == "optlang-gurobi":
        pytest.xfail("FVA currently buggy")
    # test non-fva version (these should be fixed for textbook model)
    expected_entry = ['     o2_e    21.8      19.9      23.7       h2o_e     '
                      '29.2         25       30.7     Biomass_Ecoli_core   '
                      '0.874   ']

    model.solver = opt_solver
    solution = model.optimize()

    with captured_output() as (out, _):
        print(model.summary(solution, fva=fraction))
        check_in_line(out.getvalue(), expected_entry)


@pytest.mark.parametrize("fraction", [0.95])
def test_model_summary_to_frame_with_fva(model, opt_solver, fraction):
    """Test model summary.to_frame() (using FVA)."""
    if opt_solver == "optlang-gurobi":
        pytest.xfail("FVA currently buggy")
    # test non-fva version (these should be fixed for textbook model)
    expected_in_fluxes = ['o2_e', 'glc__D_e', 'nh4_e', 'pi_e', np.nan,
                          np.nan, np.nan, np.nan, np.nan, np.nan, np.nan,
                          np.nan]
    expected_out_fluxes = ['h2o_e', 'co2_e', 'h_e', 'for_e', 'ac_e', 'acald_e',
                           'pyr_e', 'etoh_e', 'lac__D_e', 'succ_e', 'akg_e',
                           'glu__L_e']

    model.solver = opt_solver
    solution = model.optimize()
    out_df = model.summary(solution, fva=fraction).to_frame()

    assert out_df[('IN_FLUXES', 'ID')].tolist() == expected_in_fluxes
    assert out_df[('OUT_FLUXES', 'ID')].tolist() == expected_out_fluxes



# -*- coding: utf-8 -*-

"""Test functionalities of ReactionSummary."""

from __future__ import absolute_import

import numpy as np
import pytest

from cobra.test.test_core.test_summary import (
    captured_output, check_in_line, check_line)


@pytest.mark.skip(reason="has some weird formatting issue")
@pytest.mark.parametrize("rxn, names", [("ACALD", False), ("ACALD", True)])
def test_reaction_summary_to_table(model, rxn, names):
    """Test reaction summary._to_table()."""

    if names:
        expected_entry = [' adhE                                     '
                          'Coenzyme A             -1                   c    ']
    else:
        expected_entry = ['          accoa_c               '
                          '1                   c    ']

    with captured_output() as (out, _):
        print(model.reactions.get_by_id(rxn).summary(names=names))
    check_in_line(out.getvalue(), expected_entry)


@pytest.mark.parametrize("rxn, names", [("ACALD", False), ("FUM", True)])
def test_reaction_summary_to_frame(model, rxn, names):
    """Test reaction summary.to_frame()."""

    out_df = model.reactions.get_by_id(rxn).summary(names=names).to_frame()

    if names:
        expected_gene_names = ['fumB', 'fumC', 'fumA']
        expected_met_names = ['Fumarate', 'H2O', 'L-Malate']

    else:
        expected_gene_names = ['b0351', 'b1241', np.nan, np.nan, np.nan,
                               np.nan]
        expected_met_names = ['acald_c', 'coa_c', 'nad_c', 'accoa_c', 'h_c',
                              'nadh_c']

    assert all(out_df['GENES', 'ID'].tolist()) == \
        all(expected_gene_names)

    assert all(out_df['METABOLITES', 'ID'].tolist()) == \
        all(expected_met_names)



# -*- coding: utf-8 -*-

from __future__ import absolute_import

import re
import sys
from contextlib import contextmanager

from six import StringIO


@contextmanager
def captured_output():
    """A context manager to test the IO summary methods."""
    new_out, new_err = StringIO(), StringIO()
    old_out, old_err = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_out, new_err
        yield sys.stdout, sys.stderr

    finally:
        sys.stdout, sys.stderr = old_out, old_err


def check_line(output, expected_entries,
               pattern=re.compile(r"\s")):
    """Ensure each expected entry is in the output."""
    output_set = set(
        pattern.sub("", line) for line in output.splitlines())
    for elem in expected_entries:
        assert pattern.sub("", elem) in output_set


def check_in_line(output, expected_entries,
                  pattern=re.compile(r"\s")):
    """Ensure each expected entry is contained in the output."""
    output_strip = [pattern.sub("", line) for line in
                    output.splitlines()]
    for elem in expected_entries:
        assert any(
            pattern.sub("", elem) in line for line in output_strip), \
            "Not found: {} in:\n{}".format(pattern.sub("", elem),
                                           "\n".join(output_strip))



# -*- coding: utf-8 -*-

"""Test functions of dictlist.py"""

from __future__ import absolute_import

import re
from copy import copy, deepcopy
from pickle import HIGHEST_PROTOCOL, dumps, loads

import pytest
from six.moves import range

from cobra.core import DictList, Object


@pytest.fixture(scope="function")
def dict_list():
    obj = Object("test1")
    test_list = DictList()
    test_list.append(obj)
    return obj, test_list


def test_contains(dict_list):
    obj, test_list = dict_list
    assert obj in test_list
    assert obj.id in test_list
    assert Object("not_in") not in test_list
    assert 'not_in' not in test_list


def test_index(dict_list):
    obj, test_list = dict_list
    assert test_list.index("test1") == 0
    assert test_list.index(obj) == 0
    with pytest.raises(ValueError):
        test_list.index("f")
    with pytest.raises(ValueError):
        test_list.index(Object("f"))
    # Ensure indexing with an object that is a different object
    # also raises an error
    with pytest.raises(ValueError):
        test_list.index(Object("test1"))


def test_independent():
    a = DictList([Object("o1"), Object("o2")])
    b = DictList()
    assert "o1" in a
    assert "o1" not in b
    b.append(Object("o3"))
    assert "o3" not in a
    assert "o3" in b


def test_get_by_any(dict_list):
    obj, test_list = dict_list
    assert test_list.get_by_any(0) == [obj]
    assert test_list.get_by_any('test1') == [obj]
    with pytest.raises(KeyError):
        test_list.get_by_any('not-in-list')
    with pytest.raises(TypeError):
        test_list.get_by_any(1.1)
    assert test_list.get_by_any(obj) == [obj]


def test_append(dict_list):
    obj, test_list = dict_list
    obj2 = Object("test2")
    test_list.append(obj2)
    with pytest.raises(ValueError):
        test_list.append(Object("test1"))
    assert test_list.index(obj2) == 1
    assert test_list[1] == obj2
    assert test_list.get_by_id("test2") is obj2
    assert len(test_list) == 2


def test_insert(dict_list):
    obj, test_list = dict_list
    obj2 = Object("a")
    test_list.insert(0, obj2)
    assert test_list.index(obj2) == 0
    assert test_list.index("test1") == 1
    assert test_list.get_by_id("a") is obj2
    assert len(test_list) == 2
    with pytest.raises(ValueError):
        test_list.append(obj2)


def test_extend(dict_list):
    obj, test_list = dict_list
    obj_list = [Object("test%d" % (i)) for i in range(2, 10)]
    test_list.extend(obj_list)
    assert test_list[1].id == "test2"
    assert test_list.get_by_id("test2") == obj_list[0]
    assert test_list[8].id == "test9"
    assert len(test_list) == 9
    with pytest.raises(ValueError):
        test_list.extend([Object("test1")])
    # Even if the object is unique, if it is present twice in the new
    # list, it should still raise an exception
    with pytest.raises(ValueError):
        test_list.extend([Object("testd"), Object("testd")])


def test_iadd(dict_list):
    obj, test_list = dict_list
    obj_list = [Object("test%d" % (i)) for i in range(2, 10)]
    test_list += obj_list
    assert test_list[1].id == "test2"
    assert test_list.get_by_id("test2") == obj_list[0]
    assert test_list[8].id == "test9"
    assert len(test_list) == 9


def test_add(dict_list):
    obj, test_list = dict_list
    obj_list = [Object("test%d" % (i)) for i in range(2, 10)]
    sum = test_list + obj_list
    assert sum is not test_list
    assert sum is not obj_list
    assert test_list[0].id == "test1"
    assert sum[1].id == "test2"
    assert sum.get_by_id("test2") == obj_list[0]
    assert sum[8].id == "test9"
    assert len(test_list) == 1
    assert len(sum) == 9


def test_sub(dict_list):
    obj, test_list = dict_list
    obj_list = [Object("test%d" % i) for i in range(2, 10)]
    sum = test_list + obj_list
    sub = sum - test_list
    assert test_list[0].id == "test1"
    assert sub[0].id == "test2"
    assert len(sub) == 8
    assert sum - obj_list == test_list


def test_isub(dict_list):
    obj, test_list = dict_list
    obj_list = [Object("test%d" % i) for i in range(2, 10)]
    sum = test_list + obj_list
    sum -= obj_list[2:4]
    assert len(sum) == 7
    with pytest.raises(ValueError):
        sum -= [Object('bogus')]


def test_init_copy(dict_list):
    obj, test_list = dict_list
    test_list.append(Object("test2"))
    copied = DictList(test_list)
    assert test_list is not copied
    assert isinstance(copied, test_list.__class__)
    assert len(test_list) == len(copied)
    for i, v in enumerate(test_list):
        assert test_list[i].id == copied[i].id
        assert i == copied.index(v.id)
        assert test_list[i] is copied[i]
        assert v is copied.get_by_id(v.id)


def test_slice(dict_list):
    obj, test_list = dict_list
    test_list.append(Object("test2"))
    test_list.append(Object("test3"))
    sliced = test_list[:-1]
    assert test_list is not sliced
    assert isinstance(sliced, test_list.__class__)
    assert len(test_list) == len(sliced) + 1
    for i, v in enumerate(sliced):
        assert test_list[i].id == sliced[i].id
        assert i == sliced.index(v.id)
        assert test_list[i] is sliced[i]
        assert test_list[i] is sliced.get_by_id(v.id)


def test_copy(dict_list):
    obj, test_list = dict_list
    test_list.append(Object("test2"))
    copied = copy(test_list)
    assert test_list is not copied
    assert isinstance(copied, test_list.__class__)
    assert len(test_list) == len(copied)
    for i, v in enumerate(test_list):
        assert test_list[i].id == copied[i].id
        assert i == copied.index(v.id)
        assert test_list[i] is copied[i]
        assert v is copied.get_by_id(v.id)


def test_deepcopy(dict_list):
    obj, test_list = dict_list
    test_list.append(Object("test2"))
    copied = deepcopy(test_list)
    assert test_list is not copied
    assert isinstance(copied, test_list.__class__)
    assert len(test_list) == len(copied)
    for i, v in enumerate(test_list):
        assert test_list[i].id == copied[i].id
        assert i == copied.index(v.id)
        assert test_list[i] is not copied[i]
        assert v is not copied.get_by_id(v.id)


def test_pickle(dict_list):
    obj, test_list = dict_list
    test_list.append(Object("test2"))
    for protocol in range(HIGHEST_PROTOCOL):
        pickle_str = dumps(test_list, protocol=protocol)
        copied = loads(pickle_str)
        assert test_list is not copied
        assert isinstance(copied, test_list.__class__)
        assert len(test_list) == len(copied)
        for i, v in enumerate(test_list):
            assert test_list[i].id == copied[i].id
            assert i == copied.index(v.id)
            assert test_list[i] is not copied[i]
            assert v is not copied.get_by_id(v.id)


def test_query(dict_list):
    obj, test_list = dict_list
    obj2 = Object("test2")
    obj2.name = "foobar1"
    test_list.append(obj2)
    result = test_list.query("test1")  # matches only test1
    assert len(result) == 1
    result = test_list.query(u"test1", "id")  # matches with unicode
    assert len(result) == 1
    assert result[0] == obj
    result = test_list.query("foo", "name")  # matches only test2
    assert len(result) == 1
    assert result[0] == obj2
    result = test_list.query("test", "id")  # matches test1 and test2
    assert len(result) == 2
    # Test with a regular expression
    result = test_list.query(re.compile("test[0-9]"), "id")
    assert len(result) == 2
    result = test_list.query(re.compile("test[29]"), "id")
    assert len(result) == 1
    # Test query of name
    result = test_list.query(re.compile("foobar."), "name")
    assert len(result) == 1
    # Test query with lambda function
    result = test_list.query(lambda x: x.id == 'test1')
    assert len(result) == 1


def test_removal():
    obj_list = DictList(Object("test%d" % (i)) for i in range(2, 10))
    del obj_list[3]
    assert "test5" not in obj_list
    assert obj_list.index(obj_list[-1]) == len(obj_list) - 1
    assert len(obj_list) == 7
    del obj_list[3:5]
    assert "test6" not in obj_list
    assert "test7" not in obj_list
    assert obj_list.index(obj_list[-1]) == len(obj_list) - 1
    assert len(obj_list) == 5
    removed = obj_list.pop(1)
    assert obj_list.index(obj_list[-1]) == len(obj_list) - 1
    assert removed.id == "test3"
    assert "test3" not in obj_list
    assert len(obj_list) == 4
    removed = obj_list.pop()
    assert removed.id == "test9"
    assert removed.id not in obj_list
    assert len(obj_list) == 3


def test_set():
    obj_list = DictList(Object("test%d" % (i)) for i in range(10))
    obj_list[4] = Object("testa")
    assert obj_list.index("testa") == 4
    assert obj_list[4].id == "testa"
    obj_list[5:7] = [Object("testb"), Object("testc")]
    assert obj_list.index("testb") == 5
    assert obj_list[5].id == "testb"
    assert obj_list.index("testc") == 6
    assert obj_list[6].id == "testc"
    # Even if the object is unique, if it is present twice in the new
    # list, it should still raise an exception
    with pytest.raises(ValueError):
        obj_list.__setitem__(slice(5, 7),
                             [Object("testd"), Object("testd")])


def test_sort_and_reverse():
    dl = DictList(Object("test%d" % (i)) for i in reversed(range(10)))
    assert dl[0].id == "test9"
    dl.sort()
    assert len(dl) == 10
    assert dl[0].id == "test0"
    assert dl.index("test0") == 0
    dl.reverse()
    assert dl[0].id == "test9"
    assert dl.index("test0") == 9


def test_dir(dict_list):
    obj, test_list = dict_list
    # Make sure tab completion works
    attrs = dir(test_list)
    assert "test1" in attrs
    assert "_dict" in attrs  # attribute of DictList


def test_union(dict_list):
    obj, test_list = dict_list
    test_list.union([Object("test1"), Object("test2")])
    # Add only 1 element
    assert len(test_list) == 2
    assert test_list.index("test2") == 1



# -*- coding: utf-8 -*-

"""Provide base class for Hit-and-Run samplers.

New samplers should derive from the abstract `HRSampler` class
where possible to provide a uniform interface."""

from __future__ import absolute_import, division

import ctypes
from collections import namedtuple
from logging import getLogger
from multiprocessing import Array
from time import time

import numpy as np
from optlang.interface import OPTIMAL
from optlang.symbolics import Zero

from cobra.util import (
    constraint_matrices, create_stoichiometric_matrix, nullspace)


LOGGER = getLogger(__name__)


# Maximum number of retries for sampling
MAX_TRIES = 100


Problem = namedtuple("Problem",
                     ["equalities", "b", "inequalities", "bounds",
                      "variable_fixed", "variable_bounds", "nullspace",
                      "homogeneous"])
"""Defines the matrix representation of a sampling problem.

Attributes
----------
equalities : numpy.array
    All equality constraints in the model.
b : numpy.array
    The right side of the equality constraints.
inequalities : numpy.array
    All inequality constraints in the model.
bounds : numpy.array
    The lower and upper bounds for the inequality constraints.
variable_bounds : numpy.array
    The lower and upper bounds for the variables.
homogeneous: boolean
    Indicates whether the sampling problem is homogenous, e.g. whether there
    exist no non-zero fixed variables or constraints.
nullspace : numpy.matrix
    A matrix containing the nullspace of the equality constraints. Each column
    is one basis vector.

"""


def shared_np_array(shape, data=None, integer=False):
    """Create a new numpy array that resides in shared memory.

    Parameters
    ----------
    shape : tuple of ints
        The shape of the new array.
    data : numpy.array
        Data to copy to the new array. Has to have the same shape.
    integer : boolean
        Whether to use an integer array. Defaults to False which means
        float array.

    """

    size = np.prod(shape)

    if integer:
        array = Array(ctypes.c_int64, int(size))
        np_array = np.frombuffer(array.get_obj(), dtype="int64")
    else:
        array = Array(ctypes.c_double, int(size))
        np_array = np.frombuffer(array.get_obj())

    np_array = np_array.reshape(shape)

    if data is not None:
        if len(shape) != len(data.shape):
            raise ValueError("`data` must have the same dimensions"
                             "as the created array.")
        same = all(x == y for x, y in zip(shape, data.shape))

        if not same:
            raise ValueError("`data` must have the same shape"
                             "as the created array.")
        np_array[:] = data

    return np_array


class HRSampler(object):
    """The abstract base class for hit-and-run samplers.

    Parameters
    ----------
    model : cobra.Model
        The cobra model from which to generate samples.
    thinning : int
        The thinning factor of the generated sampling chain. A thinning of 10
        means samples are returned every 10 steps.
    nproj : int > 0, optional
        How often to reproject the sampling point into the feasibility space.
        Avoids numerical issues at the cost of lower sampling. If you observe
        many equality constraint violations with `sampler.validate` you should
        lower this number.
    seed : int > 0, optional
        The random number seed that should be used.

    Attributes
    ----------
    model : cobra.Model
        The cobra model from which the sampes get generated.
    feasibility_tol: float
        The tolerance used for checking equalities feasibility.
    bounds_tol: float
        The tolerance used for checking bounds feasibility.
    thinning : int
        The currently used thinning factor.
    n_samples : int
        The total number of samples that have been generated by this
        sampler instance.
    retries : int
        The overall of sampling retries the sampler has observed. Larger
        values indicate numerical instabilities.
    problem : collections.namedtuple
        A python object whose attributes define the entire sampling problem in
        matrix form. See docstring of `Problem`.
    warmup : numpy.matrix
        A matrix of with as many columns as reactions in the model and more
        than 3 rows containing a warmup sample in each row. None if no warmup
        points have been generated yet.
    nproj : int
        How often to reproject the sampling point into the feasibility space.
    seed : int > 0, optional
        Sets the random number seed. Initialized to the current time stamp if
        None.
    fwd_idx : numpy.array
        Has one entry for each reaction in the model containing the index of
        the respective forward variable.
    rev_idx : numpy.array
        Has one entry for each reaction in the model containing the index of
        the respective reverse variable.

    """

    def __init__(self, model, thinning, nproj=None, seed=None):
        """Initialize a new sampler object."""

        # This currently has to be done to reset the solver basis which is
        # required to get deterministic warmup point generation
        # (in turn required for a working `seed` arg)
        if model.solver.is_integer:
            raise TypeError("sampling does not work with integer problems :(")

        self.model = model.copy()
        self.feasibility_tol = model.tolerance
        self.bounds_tol = model.tolerance
        self.thinning = thinning

        if nproj is None:
            self.nproj = int(min(len(self.model.variables)**3, 1e6))
        else:
            self.nproj = nproj

        self.n_samples = 0
        self.retries = 0
        self.problem = self.__build_problem()

        # Set up a map from reaction -> forward/reverse variable
        var_idx = {v: idx for idx, v in enumerate(self.model.variables)}

        self.fwd_idx = np.array([var_idx[r.forward_variable]
                                 for r in self.model.reactions])
        self.rev_idx = np.array([var_idx[r.reverse_variable]
                                 for r in self.model.reactions])
        self.warmup = None

        if seed is None:
            self._seed = int(time())
        else:
            self._seed = seed

        # Avoid overflow
        self._seed = self._seed % np.iinfo(np.int32).max

    def __build_problem(self):
        """Build the matrix representation of the sampling problem."""

        # Set up the mathematical problem
        prob = constraint_matrices(self.model, zero_tol=self.feasibility_tol)

        # check if there any non-zero equality constraints
        equalities = prob.equalities
        b = prob.b
        bounds = np.atleast_2d(prob.bounds).T
        var_bounds = np.atleast_2d(prob.variable_bounds).T
        homogeneous = all(np.abs(b) < self.feasibility_tol)
        fixed_non_zero = np.abs(prob.variable_bounds[:, 1]) > \
            self.feasibility_tol
        fixed_non_zero &= prob.variable_fixed

        # check if there are any non-zero fixed variables, add them as
        # equalities to the stoichiometric matrix
        if any(fixed_non_zero):
            n_fixed = fixed_non_zero.sum()
            rows = np.zeros((n_fixed, prob.equalities.shape[1]))
            rows[range(n_fixed), np.where(fixed_non_zero)] = 1.0
            equalities = np.vstack([equalities, rows])
            var_b = prob.variable_bounds[:, 1]
            b = np.hstack([b, var_b[fixed_non_zero]])
            homogeneous = False

        # Set up a projection that can cast point into the nullspace
        nulls = nullspace(equalities)

        # convert bounds to a matrix and add variable bounds as well
        return Problem(
            equalities=shared_np_array(equalities.shape, equalities),
            b=shared_np_array(b.shape, b),
            inequalities=shared_np_array(prob.inequalities.shape,
                                         prob.inequalities),
            bounds=shared_np_array(bounds.shape, bounds),
            variable_fixed=shared_np_array(prob.variable_fixed.shape,
                                           prob.variable_fixed, integer=True),
            variable_bounds=shared_np_array(var_bounds.shape, var_bounds),
            nullspace=shared_np_array(nulls.shape, nulls),
            homogeneous=homogeneous
        )

    def generate_fva_warmup(self):
        """Generate the warmup points for the sampler.

        Generates warmup points by setting each flux as the sole objective
        and minimizing/maximizing it. Also caches the projection of the
        warmup points into the nullspace for non-homogeneous problems (only
        if necessary).

        """

        self.n_warmup = 0
        reactions = self.model.reactions
        self.warmup = np.zeros((2 * len(reactions), len(self.model.variables)))
        self.model.objective = Zero

        for sense in ("min", "max"):
            self.model.objective_direction = sense

            for i, r in enumerate(reactions):
                variables = (self.model.variables[self.fwd_idx[i]],
                             self.model.variables[self.rev_idx[i]])

                # Omit fixed reactions if they are non-homogeneous
                if r.upper_bound - r.lower_bound < self.bounds_tol:
                    LOGGER.info("skipping fixed reaction %s" % r.id)
                    continue

                self.model.objective.set_linear_coefficients(
                    {variables[0]: 1, variables[1]: -1})

                self.model.slim_optimize()

                if not self.model.solver.status == OPTIMAL:
                    LOGGER.info("can not maximize reaction %s, skipping it" %
                                r.id)
                    continue

                primals = self.model.solver.primal_values
                sol = [primals[v.name] for v in self.model.variables]
                self.warmup[self.n_warmup, ] = sol
                self.n_warmup += 1

                # Reset objective
                self.model.objective.set_linear_coefficients(
                    {variables[0]: 0, variables[1]: 0})

        # Shrink to measure
        self.warmup = self.warmup[0:self.n_warmup, :]

        # Remove redundant search directions
        keep = np.logical_not(self._is_redundant(self.warmup))
        self.warmup = self.warmup[keep, :]
        self.n_warmup = self.warmup.shape[0]

        # Catch some special cases
        if len(self.warmup.shape) == 1 or self.warmup.shape[0] == 1:
            raise ValueError("Your flux cone consists only of a single point!")
        elif self.n_warmup == 2:
            if not self.problem.homogeneous:
                raise ValueError("Can not sample from an inhomogenous problem"
                                 " with only 2 search directions :(")
            LOGGER.info("All search directions on a line, adding another one.")
            newdir = self.warmup.T.dot([0.25, 0.25])
            self.warmup = np.vstack([self.warmup, newdir])
            self.n_warmup += 1

        # Shrink warmup points to measure
        self.warmup = shared_np_array(
            (self.n_warmup, len(self.model.variables)), self.warmup)

    def _reproject(self, p):
        """Reproject a point into the feasibility region.

        This function is guaranteed to return a new feasible point. However,
        no guarantees in terms of proximity to the original point can be made.

        Parameters
        ----------
        p : numpy.array
            The current sample point.

        Returns
        -------
        numpy.array
            A new feasible point. If `p` was feasible it wil return p.

        """

        nulls = self.problem.nullspace
        equalities = self.problem.equalities

        # don't reproject if point is feasible
        if np.allclose(equalities.dot(p), self.problem.b,
                       rtol=0, atol=self.feasibility_tol):
            new = p
        else:
            LOGGER.info("feasibility violated in sample"
                        " %d, trying to reproject" % self.n_samples)
            new = nulls.dot(nulls.T.dot(p))

        # Projections may violate bounds
        # set to random point in space in that case
        if any(new != p):
            LOGGER.info("reprojection failed in sample"
                        " %d, using random point in space" % self.n_samples)
            new = self._random_point()

        return new

    def _random_point(self):
        """Find an approximately random point in the flux cone."""

        idx = np.random.randint(self.n_warmup,
                                size=min(2, np.ceil(np.sqrt(self.n_warmup))))
        return self.warmup[idx, :].mean(axis=0)

    def _is_redundant(self, matrix, cutoff=None):
        """Identify rdeundant rows in a matrix that can be removed."""

        cutoff = 1.0 - self.feasibility_tol

        # Avoid zero variances
        extra_col = matrix[:, 0] + 1

        # Avoid zero rows being correlated with constant rows
        extra_col[matrix.sum(axis=1) == 0] = 2
        corr = np.corrcoef(np.c_[matrix, extra_col])
        corr = np.tril(corr, -1)

        return (np.abs(corr) > cutoff).any(axis=1)

    def _bounds_dist(self, p):
        """Get the lower and upper bound distances. Negative is bad."""

        prob = self.problem
        lb_dist = (p - prob.variable_bounds[0, ]).min()
        ub_dist = (prob.variable_bounds[1, ] - p).min()

        if prob.bounds.shape[0] > 0:
            const = prob.inequalities.dot(p)
            const_lb_dist = (const - prob.bounds[0, ]).min()
            const_ub_dist = (prob.bounds[1, ] - const).min()
            lb_dist = min(lb_dist, const_lb_dist)
            ub_dist = min(ub_dist, const_ub_dist)

        return np.array([lb_dist, ub_dist])

    def sample(self, n, fluxes=True):
        """Abstract sampling function.

        Should be overwritten by child classes.

        """
        pass

    def batch(self, batch_size, batch_num, fluxes=True):
        """Create a batch generator.

        This is useful to generate n batches of m samples each.

        Parameters
        ----------
        batch_size : int
            The number of samples contained in each batch (m).
        batch_num : int
            The number of batches in the generator (n).
        fluxes : boolean
            Whether to return fluxes or the internal solver variables. If set
            to False will return a variable for each forward and backward flux
            as well as all additional variables you might have defined in the
            model.

        Yields
        ------
        pandas.DataFrame
            A DataFrame with dimensions (batch_size x n_r) containing
            a valid flux sample for a total of n_r reactions (or variables if
            fluxes=False) in each row.

        """

        for i in range(batch_num):
            yield self.sample(batch_size, fluxes=fluxes)

    def validate(self, samples):
        """Validate a set of samples for equality and inequality feasibility.

        Can be used to check whether the generated samples and warmup points
        are feasible.

        Parameters
        ----------
        samples : numpy.matrix
            Must be of dimension (n_samples x n_reactions). Contains the
            samples to be validated. Samples must be from fluxes.

        Returns
        -------
        numpy.array
            A one-dimensional numpy array of length containing
            a code of 1 to 3 letters denoting the validation result:

            - 'v' means feasible in bounds and equality constraints
            - 'l' means a lower bound violation
            - 'u' means a lower bound validation
            - 'e' means and equality constraint violation

        """

        samples = np.atleast_2d(samples)
        prob = self.problem

        if samples.shape[1] == len(self.model.reactions):
            S = create_stoichiometric_matrix(self.model)
            b = np.array([self.model.constraints[m.id].lb for m in
                          self.model.metabolites])
            bounds = np.array([r.bounds for r in self.model.reactions]).T
        elif samples.shape[1] == len(self.model.variables):
            S = prob.equalities
            b = prob.b
            bounds = prob.variable_bounds
        else:
            raise ValueError("Wrong number of columns. samples must have a "
                             "column for each flux or variable defined in the "
                             "model!")

        feasibility = np.abs(S.dot(samples.T).T - b).max(axis=1)
        lb_error = (samples - bounds[0, ]).min(axis=1)
        ub_error = (bounds[1, ] - samples).min(axis=1)

        if (samples.shape[1] == len(self.model.variables) and
                prob.inequalities.shape[0]):
            consts = prob.inequalities.dot(samples.T)
            lb_error = np.minimum(
                lb_error,
                (consts - prob.bounds[0, ]).min(axis=1))
            ub_error = np.minimum(
                ub_error,
                (prob.bounds[1, ] - consts).min(axis=1)
            )

        valid = (
            (feasibility < self.feasibility_tol) &
            (lb_error > -self.bounds_tol) &
            (ub_error > -self.bounds_tol))
        codes = np.repeat("", valid.shape[0]).astype(np.dtype((str, 3)))
        codes[valid] = "v"
        codes[lb_error <= -self.bounds_tol] = np.char.add(
            codes[lb_error <= -self.bounds_tol], "l")
        codes[ub_error <= -self.bounds_tol] = np.char.add(
            codes[ub_error <= -self.bounds_tol], "u")
        codes[feasibility > self.feasibility_tol] = np.char.add(
            codes[feasibility > self.feasibility_tol], "e")

        return codes


# Required by ACHRSampler and OptGPSampler
# Has to be declared outside of class to be used for multiprocessing :(
def step(sampler, x, delta, fraction=None, tries=0):
    """Sample a new feasible point from the point `x` in direction `delta`."""

    prob = sampler.problem
    valid = ((np.abs(delta) > sampler.feasibility_tol) &
             np.logical_not(prob.variable_fixed))

    # permissible alphas for staying in variable bounds
    valphas = ((1.0 - sampler.bounds_tol) * prob.variable_bounds -
               x)[:, valid]
    valphas = (valphas / delta[valid]).flatten()

    if prob.bounds.shape[0] > 0:
        # permissible alphas for staying in constraint bounds
        ineqs = prob.inequalities.dot(delta)
        valid = np.abs(ineqs) > sampler.feasibility_tol
        balphas = ((1.0 - sampler.bounds_tol) * prob.bounds -
                   prob.inequalities.dot(x))[:, valid]
        balphas = (balphas / ineqs[valid]).flatten()

        # combined alphas
        alphas = np.hstack([valphas, balphas])
    else:
        alphas = valphas
    pos_alphas = alphas[alphas > 0.0]
    neg_alphas = alphas[alphas <= 0.0]
    alpha_range = np.array([neg_alphas.max() if len(neg_alphas) > 0 else 0,
                            pos_alphas.min() if len(pos_alphas) > 0 else 0])

    if fraction:
        alpha = alpha_range[0] + fraction * (alpha_range[1] - alpha_range[0])
    else:
        alpha = np.random.uniform(alpha_range[0], alpha_range[1])

    p = x + alpha * delta

    # Numerical instabilities may cause bounds invalidation
    # reset sampler and sample from one of the original warmup directions
    # if that occurs. Also reset if we got stuck.
    if (np.any(sampler._bounds_dist(p) < -sampler.bounds_tol) or
            np.abs(np.abs(alpha_range).max() * delta).max() <
            sampler.bounds_tol):
        if tries > MAX_TRIES:
            raise RuntimeError("Can not escape sampling region, model seems"
                               " numerically unstable :( Reporting the "
                               "model to "
                               "https://github.com/opencobra/cobrapy/issues "
                               "will help us to fix this :)")
        LOGGER.info("found bounds infeasibility in sample, "
                    "resetting to center")
        newdir = sampler.warmup[np.random.randint(sampler.n_warmup)]
        sampler.retries += 1

        return step(sampler, sampler.center, newdir - sampler.center, None,
                    tries + 1)
    return p



# -*- coding: utf-8 -*-

"""Imports for the media module."""

from cobra.medium.boundary_types import (
    find_boundary_types, find_external_compartment, is_boundary_type,
    sbo_terms)
from cobra.medium.minimal_medium import minimal_medium



# -*- coding: utf-8 -*-

"""Provide OptGP sampler."""

from __future__ import absolute_import, division

from multiprocessing import Pool

import numpy as np
import pandas

from cobra.core.configuration import Configuration
from cobra.sampling.hr_sampler import HRSampler, shared_np_array, step


__all__ = ("OptGPSampler",)


CONFIGURATION = Configuration()


def mp_init(obj):
    """Initialize the multiprocessing pool."""

    global sampler
    sampler = obj


# Unfortunately this has to be outside the class to be usable with
# multiprocessing :()
def _sample_chain(args):
    """Sample a single chain for OptGPSampler.

    center and n_samples are updated locally and forgotten afterwards.

    """

    n, idx = args       # has to be this way to work in Python 2.7
    center = sampler.center
    np.random.seed((sampler._seed + idx) % np.iinfo(np.int32).max)
    pi = np.random.randint(sampler.n_warmup)

    prev = sampler.warmup[pi, ]
    prev = step(sampler, center, prev - center, 0.95)

    n_samples = max(sampler.n_samples, 1)
    samples = np.zeros((n, center.shape[0]))

    for i in range(1, sampler.thinning * n + 1):
        pi = np.random.randint(sampler.n_warmup)
        delta = sampler.warmup[pi, ] - center

        prev = step(sampler, prev, delta)

        if sampler.problem.homogeneous and (
                n_samples * sampler.thinning % sampler.nproj == 0):
            prev = sampler._reproject(prev)
            center = sampler._reproject(center)

        if i % sampler.thinning == 0:
            samples[i//sampler.thinning - 1, ] = prev

        center = ((n_samples * center) / (n_samples + 1) +
                  prev / (n_samples + 1))
        n_samples += 1

    return (sampler.retries, samples)


class OptGPSampler(HRSampler):
    """A parallel optimized sampler.

    A parallel sampler with fast convergence and parallel execution. See [1]_
    for details.

    Parameters
    ----------
    model : cobra.Model
        The cobra model from which to generate samples.
    processes: int, optional (default Configuration.processes)
        The number of processes used during sampling.
    thinning : int, optional
        The thinning factor of the generated sampling chain. A thinning of 10
        means samples are returned every 10 steps.
    nproj : int > 0, optional
        How often to reproject the sampling point into the feasibility space.
        Avoids numerical issues at the cost of lower sampling. If you observe
        many equality constraint violations with `sampler.validate` you should
        lower this number.
    seed : int > 0, optional
        Sets the random number seed. Initialized to the current time stamp if
        None.

    Attributes
    ----------
    model : cobra.Model
        The cobra model from which the samples get generated.
    thinning : int
        The currently used thinning factor.
    n_samples : int
        The total number of samples that have been generated by this
        sampler instance.
    problem : collections.namedtuple
        A python object whose attributes define the entire sampling problem in
        matrix form. See docstring of `Problem`.
    warmup : numpy.matrix
        A matrix of with as many columns as reactions in the model and more
        than 3 rows containing a warmup sample in each row. None if no warmup
        points have been generated yet.
    retries : int
        The overall of sampling retries the sampler has observed. Larger
        values indicate numerical instabilities.
    seed : int > 0, optional
        Sets the random number seed. Initialized to the current time stamp if
        None.
    nproj : int
        How often to reproject the sampling point into the feasibility space.
    fwd_idx : numpy.array
        Has one entry for each reaction in the model containing the index of
        the respective forward variable.
    rev_idx : numpy.array
        Has one entry for each reaction in the model containing the index of
        the respective reverse variable.
    prev : numpy.array
        The current/last flux sample generated.
    center : numpy.array
        The center of the sampling space as estimated by the mean of all
        previously generated samples.

    Notes
    -----
    The sampler is very similar to artificial centering where each process
    samples its own chain. Initial points are chosen randomly from the warmup
    points followed by a linear transformation that pulls the points a little
    bit towards the center of the sampling space.

    If the number of processes used is larger than the one requested,
    number of samples is adjusted to the smallest multiple of the number of
    processes larger than the requested sample number. For instance, if you
    have 3 processes and request 8 samples you will receive 9.

    Memory usage is roughly in the order of (2 * number reactions)^2
    due to the required nullspace matrices and warmup points. So large
    models easily take up a few GB of RAM. However, most of the large matrices
    are kept in shared memory. So the RAM usage is independent of the number
    of processes.

    References
    ----------
    .. [1] Megchelenbrink W, Huynen M, Marchiori E (2014)
       optGpSampler: An Improved Tool for Uniformly Sampling the Solution-Space
       of Genome-Scale Metabolic Networks.
       PLoS ONE 9(2): e86587.
       https://doi.org/10.1371/journal.pone.0086587

    """

    def __init__(self, model, processes=None, thinning=100, nproj=None,
                 seed=None):
        """Initialize a new OptGPSampler."""

        super(OptGPSampler, self).__init__(model, thinning, seed=seed)
        self.generate_fva_warmup()

        if processes is None:
            self.processes = CONFIGURATION.processes
        else:
            self.processes = processes

        # This maps our saved center into shared memory,
        # meaning they are synchronized across processes
        self.center = shared_np_array((len(model.variables), ),
                                      self.warmup.mean(axis=0))

    def sample(self, n, fluxes=True):
        """Generate a set of samples.

        This is the basic sampling function for all hit-and-run samplers.

        Parameters
        ---------
        n : int
            The minimum number of samples that are generated at once
            (see Notes).
        fluxes : boolean
            Whether to return fluxes or the internal solver variables. If set
            to False will return a variable for each forward and backward flux
            as well as all additional variables you might have defined in the
            model.

        Returns
        -------
        numpy.matrix
            Returns a matrix with `n` rows, each containing a flux sample.

        Notes
        -----
        Performance of this function linearly depends on the number
        of reactions in your model and the thinning factor.

        If the number of processes is larger than one, computation is split
        across as the CPUs of your machine. This may shorten computation time.
        However, there is also overhead in setting up parallel computation so
        we recommend to calculate large numbers of samples at once
        (`n` > 1000).

        """

        if self.processes > 1:
            n_process = np.ceil(n / self.processes).astype(int)
            n = n_process * self.processes

            # The cast to list is weird but not doing it gives recursion
            # limit errors, something weird going on with multiprocessing
            args = list(zip(
                [n_process] * self.processes, range(self.processes)))

            # No with statement or starmap here since Python 2.x
            # does not support it :(
            mp = Pool(self.processes, initializer=mp_init, initargs=(self,))
            results = mp.map(_sample_chain, args, chunksize=1)
            mp.close()
            mp.join()

            chains = np.vstack([r[1] for r in results])
            self.retries += sum(r[0] for r in results)
        else:
            mp_init(self)
            results = _sample_chain((n, 0))
            chains = results[1]

        # Update the global center
        self.center = (self.n_samples * self.center +
                       np.atleast_2d(chains).sum(0)) / (self.n_samples + n)
        self.n_samples += n

        if fluxes:
            names = [r.id for r in self.model.reactions]

            return pandas.DataFrame(
                chains[:, self.fwd_idx] - chains[:, self.rev_idx],
                columns=names)
        else:
            names = [v.name for v in self.model.variables]

            return pandas.DataFrame(chains, columns=names)

    # Models can be large so don't pass them around during multiprocessing
    def __getstate__(self):
        """Return the object for serialization."""
        d = dict(self.__dict__)
        del d['model']
        return d



# -*- coding: utf-8 -*-

"""Define the global configuration."""

from __future__ import absolute_import

import logging
import types
from multiprocessing import cpu_count
from warnings import warn

from six import string_types, with_metaclass

from cobra.core.singleton import Singleton
from cobra.exceptions import SolverNotFound
from cobra.util.solver import interface_to_str, solvers


__all__ = ("Configuration",)


LOGGER = logging.getLogger(__name__)


class BaseConfiguration(object):
    """
    Define global configuration value that will be honored by cobra functions.

    This object sets default values for the modifiable attributes like
    default solver, reaction bounds etc.

    Attributes
    ----------
    solver : {"glpk", "cplex", "gurobi"}
        The default solver for new models. The solver choices are the ones
        provided by `optlang` and solvers installed in your environment.
    tolerance: float
        The default tolerance for the solver being used (default 1E-09).
    lower_bound : float
        The standard lower bound for reversible reactions (default -1000).
    upper_bound : float
        The standard upper bound for all reactions (default 1000).
    bounds : tuple of floats
        The default reaction bounds for newly created reactions. The bounds
        are in the form of lower_bound, upper_bound (default -1000.0, 1000.0).
    processes : int
        A default number of processes to use where multiprocessing is
        possible. The default number corresponds to the number of available
        cores (hyperthreads).

    """

    def __init__(self):
        self._solver = None
        self.tolerance = 1E-07
        self.lower_bound = None
        self.upper_bound = None

        # Set the default solver from a preferred order.
        for name in ["gurobi", "cplex", "glpk"]:
            try:
                self.solver = name
            except SolverNotFound:
                continue
            else:
                break

        self.bounds = -1000.0, 1000.0

        try:
            self.processes = cpu_count()
        except NotImplementedError:
            LOGGER.warning(
                "The number of cores could not be detected - assuming 1.")
            self.processes = 1

    @property
    def solver(self):
        return self._solver

    @solver.setter
    def solver(self, value):
        not_valid_interface = SolverNotFound(
            "'{}' is not a valid solver interface. Pick one from {}.".format(
                value, ", ".join(list(solvers))))
        if isinstance(value, string_types):
            if value not in solvers:
                raise not_valid_interface
            interface = solvers[value]
        elif isinstance(value, types.ModuleType) and hasattr(value, 'Model'):
            interface = value
        else:
            raise not_valid_interface
        self._solver = interface

    @property
    def bounds(self):
        return self.lower_bound, self.upper_bound

    @bounds.setter
    def bounds(self, bounds):
        # TODO: We should consider allowing `None` for free bounds.
        assert bounds[0] <= bounds[1]
        self.lower_bound = bounds[0]
        self.upper_bound = bounds[1]

    def __repr__(self):
        return """
        solver: {solver}
        solver tolerance: {tolerance}
        lower_bound: {lower_bound}
        upper_bound: {upper_bound}
        processes: {processes}""".format(
            solver=interface_to_str(self.solver),
            tolerance=self.tolerance,
            lower_bound=self.lower_bound,
            upper_bound=self.upper_bound,
            processes=self.processes,
        )

    def _repr_html_(self):
        return """
        <table>
            <tr>
                <td><strong>Solver</strong></td>
                <td>{solver}</td>
            </tr>
            <tr>
                <td><strong>Solver tolerance</strong></td>
                <td>{tolerance}</td>
            </tr>
            <tr>
                <td><strong>Lower bound</strong></td>
                <td>{lower_bound}</td>
            </tr>
            <tr>
                <td><strong>Upper bound</strong></td>
                <td>{upper_bound}</td>
            </tr>
            <tr>
                <td><strong>Processes</strong></td>
                <td>{processes}</td>
            </tr>
        </table>""".format(
            solver=interface_to_str(self.solver),
            tolerance=self.tolerance,
            lower_bound=self.lower_bound,
            upper_bound=self.upper_bound,
            processes=self.processes,
        )


class Configuration(with_metaclass(Singleton, BaseConfiguration)):
    """Define the configuration to be singleton based."""

    pass
